# -*- mode: python; coding: utf-8 -*-
# Copyright 2020-2021 the .NET Foundation
# Licensed under the MIT License.

"""Entrypoint for the "wwtdatatool" command-line interface.

"""
import argparse
import os.path
import sys


__all__ = [
    'GLOB_PATHS_INTERNALLY',
    'EnsureGlobsExpandedAction',
    'entrypoint',
    'serve_getparser',
]


# General CLI utilities

def die(msg):
    print('error:', msg, file=sys.stderr)
    sys.exit(1)

def warn(msg):
    print('warning:', msg, file=sys.stderr)

GLOB_PATHS_INTERNALLY = (os.name == 'nt')  # non-Windows has reasonable shells

class EnsureGlobsExpandedAction(argparse.Action):
    """
    An action to handle globbing for path-list arguments on Windows.

    If the CLI program is being run from the Windows command prompt, there is no
    expansion of globs by the shell. It turns out that to get globbing behavior,
    we have to manually implement it ourselves. This class helps make this
    convenient. It can be used as a ``action=`` keyword argument to
    ``ArgumentParser.add_argument()`` and, on Windows, will process argument
    text to apply globs. This argument should generally be used on arguments
    with a ``nargs='+'`` cardinality.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, self.expand_globs(values))

    def expand_globs(self, path_args):
        if not GLOB_PATHS_INTERNALLY:
            return path_args

        import glob

        result = []

        for p in path_args:
            if not glob.has_magic(p):
                result.append(p)
            else:
                matches = glob.glob(p, recursive=True)

                if matches:
                    result += matches
                else:
                    # Unix-like behavior: unmatched glob expression is passed on through
                    result.append(p)

        return result


# "cabinet" subcommand

def cabinet_getparser(parser):
    subparsers = parser.add_subparsers(dest='cabinet_command')

    p = subparsers.add_parser('list')
    p.add_argument(
        'path',
        metavar = 'PATH',
        help = 'The path to a cabinet file.',
    )

    p = subparsers.add_parser('pack')
    p.add_argument(
        'cab_path',
        metavar = 'PATH',
        help = 'The path of the cabinet file to create.',
    )
    p.add_argument(
        'input_paths',
        nargs = '+',
        action = EnsureGlobsExpandedAction,
        metavar = 'PATHS',
        help = 'Paths to files to put into the cabinet.',
    )

    p = subparsers.add_parser('unpack')
    p.add_argument(
        'path',
        metavar = 'PATH',
        help = 'The path to a cabinet file.',
    )


def cabinet_list(settings):
    from .filecabinet import FileCabinetReader

    with open(settings.path, 'rb') as f:
        reader =  FileCabinetReader(f)

        for fn in reader.filenames():
            print(fn)


def cabinet_pack(settings):
    from .filecabinet import FileCabinetWriter
    import os.path

    writer = FileCabinetWriter()

    for fn in settings.input_paths:
        with open(fn, 'rb') as f:
            data = f.read()

        # TODO: smarter splitting
        pieces = fn.split(os.path.sep)

        for p in pieces:
            if p in ('.', '..', ''):
                die(f'illegal input path "{fn}": must be relative with no ".", ".." components')

        writer.add_file_with_data('\\'.join(pieces), data)

    with open(settings.cab_path, 'wb') as f_out:
        writer.emit(f_out)


def cabinet_unpack(settings):
    from .filecabinet import FileCabinetReader
    from os import makedirs
    from os.path import join

    with open(settings.path, 'rb') as f_in:
        reader =  FileCabinetReader(f_in)

        for fn in reader.filenames():
            data = reader.read_file(fn)
            pieces = fn.split('\\')  # paths are Windows-style

            # At least the MakeDataCabinetFile tool creates a file whose
            # paths all begin with \. We are not gonna treat those as
            # absolute paths or anything like that.
            if not len(pieces[0]):
                pieces = pieces[1:]

            if len(pieces) > 1:
                makedirs(join(*pieces[:-1]), exist_ok=True)

            with open(join(*pieces), 'wb') as f_out:
                f_out.write(data)


def cabinet_impl(settings):
    if settings.cabinet_command is None:
        print('Run the "cabinet" command with `--help` for help on its subcommands')
        return

    if settings.cabinet_command == 'list':
        return cabinet_list(settings)
    elif settings.cabinet_command == 'pack':
        return cabinet_pack(settings)
    elif settings.cabinet_command == 'unpack':
        return cabinet_unpack(settings)
    else:
        die('unrecognized "cabinet" subcommand ' + settings.cabinet_command)


# "preview" subcommand

def preview_getparser(parser):
    parser.add_argument(
        'wtml_path',
        metavar = 'PATH',
        help = 'The path to the WTML file to preview.',
    )


def preview_impl(settings):
    from .server import preview_wtml
    preview_wtml(settings.wtml_path)


# "serve" subcommand

def serve_getparser(parser):
    parser.add_argument(
        '--port',
        '-p',
        metavar = 'PORT',
        type = int,
        default = 8080,
        help = 'The port on which to listen for connections.'
    )
    parser.add_argument(
        'root_dir',
        metavar = 'PATH',
        default = '.',
        help = 'The path to the base directory of the server.',
    )


def serve_impl(settings):
    from .server import run_server
    run_server(settings)


# "tree" subcommand

def tree_getparser(parser):
    subparsers = parser.add_subparsers(dest='tree_command')

    p = subparsers.add_parser('fetch')
    p.add_argument(
        'root_url',
        metavar = 'URL',
        help = 'The URL of the initial WTML file to download.',
    )

    p = subparsers.add_parser('print-dem-urls')
    p = subparsers.add_parser('print-image-urls')
    p = subparsers.add_parser('summarize')


def tree_impl(settings):
    if settings.tree_command is None:
        print('Run the "tree" command with `--help` for help on its subcommands')
        return

    if settings.tree_command == 'fetch':
        return tree_fetch(settings)
    elif settings.tree_command == 'print-dem-urls':
        return tree_print_dem_urls(settings)
    elif settings.tree_command == 'print-image-urls':
        return tree_print_image_urls(settings)
    elif settings.tree_command == 'summarize':
        return tree_summarize(settings)
    else:
        die('unrecognized "tree" subcommand ' + settings.tree_command)


def tree_fetch(settings):
    from .folder import fetch_folder_tree

    def on_fetch(url):
        print('Fetching', url, '...')

    fetch_folder_tree(settings.root_url, '.', on_fetch)


def tree_print_dem_urls(settings):
    from .folder import Folder, walk_cached_folder_tree
    from .imageset import ImageSet
    from .place import Place

    done_urls = set()

    for treepath, item in walk_cached_folder_tree('.'):
        imgset = None

        if isinstance(item, ImageSet):
            imgset = item
        elif isinstance(item, Place):
            imgset = item.as_imageset()

        if imgset is None:
            continue

        if not imgset.dem_url or imgset.dem_url in done_urls:
            continue

        done_urls.add(imgset.dem_url)
        print(imgset.dem_url, imgset.name)


def tree_print_image_urls(settings):
    from .folder import Folder, walk_cached_folder_tree
    from .imageset import ImageSet
    from .place import Place

    done_urls = set()

    for treepath, item in walk_cached_folder_tree('.'):
        imgset = None

        if isinstance(item, ImageSet):
            imgset = item
        elif isinstance(item, Place):
            imgset = item.as_imageset()

        if imgset is None:
            continue

        for url, tag in zip((imgset.url, imgset.alt_url), ('', ' (alt)')):
            if not url or url in done_urls:
                continue

            done_urls.add(url)
            print(url, imgset.name + tag)


def tree_summarize(settings):
    from .folder import Folder, walk_cached_folder_tree
    from .imageset import ImageSet
    from .place import Place

    for treepath, item in walk_cached_folder_tree('.'):
        pfx = '  ' * len(treepath)

        if isinstance(item, Folder):
            print(pfx + 'Folder', item.name)
        elif isinstance(item, ImageSet):
            index = treepath[-1]
            print(f'{pfx}{index:03d}', 'ImageSet:', item.name, '@', item.url)
        elif isinstance(item, Place):
            maybe_imgset = item.as_imageset()
            if maybe_imgset is not None:
                index = treepath[-1]
                print(f'{pfx}{index:03d}', 'Place+ImgSet:', item.name, '@', maybe_imgset.url)


# "wtml" subcommand

def wtml_getparser(parser):
    subparsers = parser.add_subparsers(dest='wtml_command')

    p = subparsers.add_parser('merge')
    p.add_argument(
        '--merged-name',
        default = 'Folder',
        help = 'The name to give to the merged folder.',
    )
    p.add_argument(
        '--merged-thumb-url',
        default = '',
        help = 'The thumbnail URL to give to the merged folder.',
    )
    p.add_argument(
        'in_paths',
        nargs = '+',
        action = EnsureGlobsExpandedAction,
        metavar = 'IN-WTML-PATH',
        help = 'The path to the input WTML files.',
    )
    p.add_argument(
        'out_path',
        metavar = 'OUT-WTML-PATH',
        help = 'The path to the output WTML file.',
    )

    p = subparsers.add_parser('report')
    p.add_argument(
        'path',
        metavar = 'WTML',
        help = 'The path to a WTML file.',
    )

    p = subparsers.add_parser('rewrite-disk')
    p.add_argument(
        'in_path',
        metavar = 'INPUT-WTML',
        help = 'The path to the input WTML file.',
    )
    p.add_argument(
        'out_path',
        metavar = 'OUTPUT-WTML',
        help = 'The path of the rewritten, output WTML file.',
    )

    p = subparsers.add_parser('rewrite-urls')
    p.add_argument(
        'in_path',
        metavar = 'INPUT-WTML',
        help = 'The path to the input WTML file.',
    )
    p.add_argument(
        'baseurl',
        metavar = 'BASE-URL',
        help = 'The new base URL to use in the file\'s contents',
    )
    p.add_argument(
        'out_path',
        metavar = 'OUTPUT-WTML',
        help = 'The path of the rewritten, output WTML file.',
    )

    p = subparsers.add_parser('transfer-astrometry')
    p.add_argument(
        'in_path',
        metavar = 'INPUT-WTML',
        help = 'The path to the input WTML file with refined astrometric solutions.',
    )
    p.add_argument(
        'update_paths',
        nargs = '+',
        action = EnsureGlobsExpandedAction,
        metavar = 'UPDATE-WTML',
        help = 'Paths of WTML files to update with data from the input file.',
    )


def wtml_impl(settings):
    if settings.wtml_command is None:
        print('Run the "wtml" command with `--help` for help on its subcommands')
        return

    if settings.wtml_command == 'merge':
        return wtml_merge(settings)
    elif settings.wtml_command == 'report':
        return wtml_report(settings)
    elif settings.wtml_command == 'rewrite-disk':
        return wtml_rewrite_disk(settings)
    elif settings.wtml_command == 'rewrite-urls':
        return wtml_rewrite_urls(settings)
    elif settings.wtml_command == 'transfer-astrometry':
        return wtml_transfer_astrometry(settings)
    else:
        die('unrecognized "wtml" subcommand ' + settings.wtml_command)


def wtml_merge(settings):
    from urllib.parse import urljoin, urlsplit
    from .folder import Folder

    out_folder = Folder()
    out_folder.name = settings.merged_name
    out_folder.thumbnail = settings.merged_thumb_url

    rel_base = os.path.dirname(settings.out_path)

    for path in settings.in_paths:
        in_folder = Folder.from_file(path)
        cur_base_url = path.replace(os.path.sep, '/')

        def mutator(url):
            if not url:
                return url
            if urlsplit(url).netloc:
                return url  # this URL is absolute

            # Resolve this relative URL, using the path of the source WTML
            # as the basis.
            url = urljoin(cur_base_url, url)

            # Now go back to filesystem-path land, so that we can use relpath to
            # compute the new path relative to the merged folder file.
            rel = os.path.relpath(url.replace('/', os.path.sep), rel_base)

            # Finally, re-express that as a URL
            return rel.replace(os.path.sep, '/')

        in_folder.mutate_urls(mutator)
        out_folder.children += in_folder.children

    with open(settings.out_path, 'wt', encoding='utf8') as f_out:
        out_folder.write_xml(f_out)


def wtml_report(settings):
    """
    Analyze a WTML file, expect to contain a single place/imageset, and report
    its metadata contents.
    """
    from bs4 import BeautifulSoup
    from datetime import datetime
    import json
    import textwrap
    from .folder import Folder
    from .imageset import ImageSet
    from .place import Place

    f = Folder.from_file(settings.path)

    warnings_hack = [0]

    def mywarn(*args, **kwargs):
        warn(*args, **kwargs)
        warnings_hack[0] += 1

    if len(f.children) != 1:
        mywarn(f'expected WTML file to contain exactly one item; found {len(f.children)}')

    if len(f.children) == 0:
        die('cannot proceed if WTML has zero items')

    pl = f.children[0]

    if isinstance(pl, ImageSet):
        die('sorry, this program is too dumb to handle top-level imagesets right now. File a bug!')
    if not isinstance(pl, Place):
        die(f'the WTML item must be a Place; found: {c}')

    if pl.foreground_image_set is not None:
        imgset = pl.foreground_image_set
    else:
        die('the WTML Place must contain a <ForegroundImageSet> item')

    # Name:

    f_name = f.name
    p_name = pl.name
    i_name = imgset.name

    if f_name != p_name:
        mywarn(f'name of folder ({f_name}) and name of Place ({p_name}) disagree')
    if f_name != i_name:
        mywarn(f'name of folder ({f_name}) and name of ImageSet ({i_name}) disagree')

    # Our extended metadata -- needs documentation!

    channel_name = None
    item_id = None
    published8601 = None

    if pl.annotation:
        try:
            anno_data = json.loads(pl.annotation)
        except Exception as e:
            mywarn(f'Place annotation data is not valid JSON; the text is: {pl.annotation!r}')
        else:
            channel_name = anno_data.get('channel')
            item_id = anno_data.get('itemid')
            published8601 = anno_data.get('publishedUTCISO8601')
    else:
        mywarn('Place contains no Annotation metadata')

    if channel_name is None:
        mywarn('Place Annotation metadata does not contain a channel name')
        channel_report = '(none specified)'
    else:
        channel_report = channel_name

    if item_id is None:
        mywarn('Place Annotation metadata does not contain an itemid')
        item_id_report = '(none specified)'
    else:
        item_id_report = item_id

    pubdate = None

    if published8601 is not None:
        try:
            pubdate = datetime.fromisoformat(published8601)
        except Error as e:
            mywarn('publication date in Place Annotation data does not seem to be in ISO8601 format')
        else:
            if pubdate.tzinfo is None:
                mywarn('publication date does not contain timezone information')

    if pubdate is None:
        mywarn('Place Annotation metadata does not contain a valid publication date')
        pubdate_report = '(unspecified)'
    else:
        pubdate_report = pubdate

    # Text entries

    def process_html(text):
        parsed = BeautifulSoup(text, 'html.parser')
        plain_report = textwrap.wrap(
            parsed.text,
            break_long_words = False,
            break_on_hyphens = False,
        )

        tag_report = []

        for line in parsed.prettify().splitlines():
            # Determine indent for mo' pretty
            i = 0
            while i < len(line) and line[i] == ' ':
                i += 1
            indent = line[:i]

            tag_report += textwrap.wrap(
                line,
                initial_indent = indent,
                subsequent_indent = indent,
                break_long_words = False,
                break_on_hyphens = False,
            )

        return plain_report, tag_report

    if not pl.description:
        mywarn('Place has no Description')
        desc_plain_report = desc_tag_report = ['(none)']
    else:
        desc_plain_report, desc_tag_report = process_html(pl.description)

    if not imgset.credits:
        mywarn('ImageSet has no credits')
        credits_plain_report = credits_tag_report = ['(none)']
    else:
        credits_plain_report, credits_tag_report = process_html(imgset.credits)

    if not imgset.credits_url:
        mywarn('ImageSet has no CreditsUrl')
        credits_url_report = '(none)'
    else:
        credits_url_report = imgset.credits_url

    # Finally, report out

    print(f'Filename: {settings.path}')
    print(f'Title (no HTML allowed): {f_name}')
    print(f'Source channel: {channel_report}')
    print(f'Source/credit URL: {credits_url_report}')
    print(f'Item ID (should be unique within channel): {item_id_report}')
    print(f'Publication date: {pubdate_report}')
    print()

    print('Description reduced to plain text:')
    print()
    for line in desc_plain_report:
        print('   ', line)
    print()
    print('Full HTML description (check links and tags!):')
    print()
    for line in desc_tag_report:
        print('   ', line)
    print()
    print('Credits reduced to plain text:')
    print()
    for line in credits_plain_report:
        print('   ', line)
    print()
    print('Full HTML credits (check links and tags!):')
    print()
    for line in credits_tag_report:
        print('   ', line)

    print()
    n_warnings = warnings_hack[0]

    if n_warnings:
        print(f'Summary: {n_warnings} were flagged')
    else:
        print('Summary: file structure looks OK! Check description and credits HTML.')


def wtml_rewrite_disk(settings):
    from .folder import Folder, make_filesystem_url_mutator

    # Note that data URLs should be relative to the *source* WTML, which is why
    # we're basing against in_path, not out_path.
    rootdir = os.path.abspath(os.path.dirname(settings.in_path))
    mutator = make_filesystem_url_mutator(rootdir)

    f = Folder.from_file(settings.in_path)
    f.mutate_urls(mutator)

    with open(settings.out_path, 'wt', encoding='utf8') as f_out:
        f.write_xml(f_out)


def wtml_rewrite_urls(settings):
    from .folder import Folder, make_absolutizing_url_mutator

    f = Folder.from_file(settings.in_path)
    f.mutate_urls(make_absolutizing_url_mutator(settings.baseurl))

    with open(settings.out_path, 'wt', encoding='utf8') as f_out:
        f.write_xml(f_out)


def wtml_transfer_astrometry(settings):
    from .folder import Folder
    from .imageset import ImageSet
    from .place import Place

    # Tables of preferred entries ...

    places = {}
    imagesets = {}

    def add_imageset(imgset):
        if imgset is None:
            return  # convenience for Place handling

        if imgset.name in imagesets:
            print('note: imageset name "%s" appears repeatedly in input file "%s"' % (imgset.name, settings.in_path))
        else:
            imagesets[imgset.name] = imgset

    def add_place(place):
        if place.name in places:
            print('note: place name "%s" appears repeatedly in input file "%s"' % (place.name, settings.in_path))
        else:
            places[place.name] = place

        add_imageset(place.image_set)
        add_imageset(place.background_image_set)
        add_imageset(place.foreground_image_set)

    IMAGESET_ASTROMETRIC_ATTRS = [
        'data_set_type',
        'width_factor',
        'reference_frame',
        'base_degrees_per_tile',
        'projection',
        'center_x',
        'center_y',
        'offset_x',
        'offset_y',
        'rotation_deg',
    ]

    PLACE_ASTROMETRIC_ATTRS = [
        'data_set_type',
        'ra_hr',
        'dec_deg',
        'latitude',
        'longitude',
        'distance',
        'angular_size',
        'zoom_level',
        'rotation_deg',
        'angle',
        'dome_alt',
        'dome_az'
    ]

    def update_imageset(imgset):
        if imgset is None:
            return 0

        ref = imagesets.get(imgset.name)
        if ref is None:
            return 0

        for att in IMAGESET_ASTROMETRIC_ATTRS:
            setattr(imgset, att, getattr(ref, att))

        return 1

    def update_place(place):
        n_updates = 0

        ref = places.get(place.name)
        if ref is not None:
            for att in PLACE_ASTROMETRIC_ATTRS:
                setattr(place, att, getattr(ref, att))

            n_updates += 1

        n_updates += update_imageset(place.image_set)
        n_updates += update_imageset(place.background_image_set)
        n_updates += update_imageset(place.foreground_image_set)
        return n_updates

    # Load up the preferred data

    in_folder = Folder.from_file(settings.in_path)

    for depth, path, item in in_folder.walk(download=False):
        if isinstance(item, Place):
            add_place(item)
        elif isinstance(item, ImageSet):
            add_imageset(item)

    # Now update everything

    n_updated_files = 0

    for update_path in settings.update_paths:
        folder = Folder.from_file(update_path)
        n_updates = 0

        for depth, path, item in folder.walk(download=False):
            if isinstance(item, Place):
                n_updates += update_place(item)
            elif isinstance(item, ImageSet):
                n_updates += update_imageset(item)

        print('%s: updated %d items' % (update_path, n_updates))

        if n_updates > 0:
            n_updated_files += 1

            with open(update_path, 'wt', encoding='utf8') as f_out:
                folder.write_xml(f_out)

    print()
    print('Updated %d WTML files.' % n_updated_files)


# The CLI driver:

def entrypoint(args=None):
    """The entrypoint for the \"wwtdatatool\" command-line interface.

    Parameters
    ----------
    args : iterable of str, or None (the default)
      The arguments on the command line. The first argument should be
      a subcommand name or global option; there is no ``argv[0]``
      parameter.

    """
    # Set up the subcommands from globals()

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subcommand")
    commands = set()

    for py_name, value in globals().items():
        if py_name.endswith('_getparser'):
            cmd_name = py_name[:-10].replace('_', '-')
            subparser = subparsers.add_parser(cmd_name)
            value(subparser)
            commands.add(cmd_name)

    # What did we get?

    settings = parser.parse_args(args)

    if settings.subcommand is None:
        print('Run me with --help for help. Allowed subcommands are:')
        print()
        for cmd in sorted(commands):
            print('   ', cmd)
        return

    py_name = settings.subcommand.replace('-', '_')

    impl = globals().get(py_name + '_impl')
    if impl is None:
        die('no such subcommand "{}"'.format(settings.subcommand))

    # OK to go!

    impl(settings)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import os.path
import shutil
import tempfile
import pytest

from .. import cli, filecabinet
from . import work_in_tempdir


def test_writer():
    with pytest.raises(ValueError):
        fcw = filecabinet.FileCabinetWriter()
        fcw.add_file_with_data('bad-file-not-bytes', u'This is not bytes.')


def test_cli(work_in_tempdir):
    "Simple smoke test to see if it runs at all."

    with open('file1.txt', 'wt', encoding='utf8') as f:
        print('Hello world', file=f)

    cli.entrypoint(['cabinet', 'pack', 'cabinet.wwtl', 'file1.txt'])

    os.remove('file1.txt')
    cli.entrypoint(['cabinet', 'unpack', 'cabinet.wwtl'])


def test_pack_globs(work_in_tempdir):
    """
    This doesn't quite belong here -- test that we properly glob filenames when
    run through the shell. This is something to worry about on Windows, when
    running in the basic command prompt.
    """
    cli.GLOB_PATHS_INTERNALLY = True

    with open('file1.txt', 'wt', encoding='utf8') as f:
        print('Hello world', file=f)

    with open('file2.txt', 'wt', encoding='utf8') as f:
        print('Hello again', file=f)

    cli.entrypoint(['cabinet', 'pack', 'cabinet1.wwtl', 'file?.txt'])
    cli.entrypoint(['cabinet', 'pack', 'cabinet2.wwtl', '?ile*.txt'])

    with open('cabinet1.wwtl', 'rb') as f:
        data1 = f.read()

    with open('cabinet2.wwtl', 'rb') as f:
        data2 = f.read()

    assert data1 == data2

    # Globby paths yield OSError, "invalid argument", on Windows.
    with pytest.raises((FileNotFoundError, OSError)):
        cli.entrypoint(['cabinet', 'pack', 'cabinet1.wwtl', '*unmatched*.txt'])



# -*- mode: python; coding: utf-8 -*-
# Copyright 2021 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import os.path

from .. import cli, enums, folder, imageset, place
from . import test_path, work_in_tempdir


def test_transfer_cli(work_in_tempdir):
    # Reference data

    iref = imageset.ImageSet()
    iref.name = 'my imageset'
    iref.data_set_type = enums.DataSetType.SKY
    iref.url = 'ref'
    iref.width_factor = 2
    iref.tile_levels = 5
    iref.base_degrees_per_tile = 1.7
    iref.projection = enums.ProjectionType.TAN
    iref.center_x = 11.11
    iref.center_y = 22.22
    iref.offset_x = 33.11
    iref.offset_y = 44.11
    iref.rotation_deg = 98.76

    pref = place.Place()
    pref.name = 'my place'
    pref.data_set_type = enums.DataSetType.SKY
    pref.ra_hr = 12.34
    pref.dec_deg = -88.88
    pref.zoom_level = 17.02
    pref.rotation_deg = 45.0
    pref.description = 'ref'
    pref.thumbnail = 'ref'
    pref.foreground_image_set = iref

    fref = folder.Folder()
    fref.children.append(pref)

    with open('reference.wtml', 'wt', encoding='utf8') as f:
        fref.write_xml(f)

    # "Updated data"

    iupd = imageset.ImageSet.from_text(iref.to_xml_string())
    # should change:
    iupd.base_degrees_per_tile = 0.0
    iupd.center_x = 0.0
    iupd.center_y = 0.0
    iupd.offset_x = 0.0
    iupd.offset_y = 0.0
    # should be preserved:
    iupd.url = 'upd-url'
    iupd.dem_url = 'upd-dem'
    iupd.mean_radius = 5555.5

    pupd = place.Place.from_text(pref.to_xml_string())
    pupd.foreground_image_set = iupd
    # should change:
    pupd.ra_hr = 0.0
    pupd.dec_deg = 0.0
    pupd.zoom_level = 0.0
    pupd.rotation_deg = 0.0
    # should be preserved:
    pupd.description = 'upd-desc'
    pupd.thumbnail = 'upd-th'

    fupd1 = folder.Folder()
    f2 = folder.Folder()
    fupd1.children.append(f2)
    f2.children.append(pupd)

    with open('update1.wtml', 'wt', encoding='utf8') as f:
        fupd1.write_xml(f)

    fupd2 = folder.Folder()
    fupd2.children.append(iupd)

    with open('update2.wtml', 'wt', encoding='utf8') as f:
        fupd2.write_xml(f)

    # Run the command

    cli.entrypoint([
        'wtml', 'transfer-astrometry',
        'reference.wtml',
        'update1.wtml', 'update2.wtml',
    ])

    # How did we do?

    iupd.base_degrees_per_tile = iref.base_degrees_per_tile
    iupd.center_x = iref.center_x
    iupd.center_y = iref.center_y
    iupd.offset_x = iref.offset_x
    iupd.offset_y = iref.offset_y

    pupd.ra_hr = pref.ra_hr
    pupd.dec_deg = pref.dec_deg
    pupd.zoom_level = pref.zoom_level
    pupd.rotation_deg = pref.rotation_deg

    u1 = folder.Folder.from_file('update1.wtml')
    assert fupd1.to_xml_string() == u1.to_xml_string()

    u2 = folder.Folder.from_file('update2.wtml')
    assert fupd2.to_xml_string() == u2.to_xml_string()



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2020 the .NET Foundation
# Licensed under the MIT License.

"""A place that a WWT user can visit.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
Place
'''.split()

from argparse import Namespace
from traitlets import Float, Instance, Int, Unicode, UseEnum

from . import LockedXmlTraits, XmlSer
from .abcs import UrlContainer
from .enums import Classification, Constellation, DataSetType
from .imageset import ImageSet

class Place(LockedXmlTraits, UrlContainer):
    """A place that can be visited."""

    data_set_type = UseEnum(
        DataSetType,
        default_value = DataSetType.EARTH
    ).tag(xml=XmlSer.attr('DataSetType'))
    name = Unicode('').tag(xml=XmlSer.attr('Name'))
    ra_hr = Float(0.0).tag(xml=XmlSer.attr('RA'))
    dec_deg = Float(0.0).tag(xml=XmlSer.attr('Dec'))
    latitude = Float(0.0).tag(xml=XmlSer.attr('Lat'))
    longitude = Float(0.0).tag(xml=XmlSer.attr('Lng'))
    constellation = UseEnum(
        Constellation,
        default_value = Constellation.UNSPECIFIED
    ).tag(xml=XmlSer.attr('Constellation'))
    classification = UseEnum(
        Classification,
        default_value = Classification.UNSPECIFIED
    ).tag(xml=XmlSer.attr('Classification'))
    magnitude = Float(0.0).tag(xml=XmlSer.attr('Magnitude'))
    distance = Float(0.0).tag(xml=XmlSer.attr('Distance'))
    angular_size = Float(0.0).tag(xml=XmlSer.attr('AngularSize'))
    zoom_level = Float(0.0).tag(xml=XmlSer.attr('ZoomLevel'))
    rotation_deg = Float(0.0).tag(xml=XmlSer.attr('Rotation'))
    angle = Float(0.0).tag(xml=XmlSer.attr('Angle'))
    opacity = Float(100.0).tag(xml=XmlSer.attr('Opacity'))
    dome_alt = Float(0.0).tag(xml=XmlSer.attr('DomeAlt'))
    dome_az = Float(0.0).tag(xml=XmlSer.attr('DomeAz'))
    background_image_set = Instance(ImageSet, allow_none=True).tag(xml=XmlSer.wrapped_inner('BackgroundImageSet'))
    foreground_image_set = Instance(ImageSet, allow_none=True).tag(xml=XmlSer.wrapped_inner('ForegroundImageSet'))
    image_set = Instance(ImageSet, allow_none=True).tag(xml=XmlSer.inner('ImageSet'))
    thumbnail = Unicode('').tag(xml=XmlSer.attr('Thumbnail'))

    description = Unicode('').tag(xml=XmlSer.text_elem('Description'))
    """
    A description of the place, using HTML markup.

    This field is not actually used in the stock WWT clients, but it is wired up
    and loaded from the XML.
    """

    annotation = Unicode('').tag(xml=XmlSer.attr('Annotation'))
    """
    Annotation metadata for the place.

    This field is only used in the web engine and web client app. The web client
    app expects this field to contain a comma-separated list of key-value pairs,
    where each pair is delimited with colons:

    .. code-block::

        key1:val1,key2:val2,key3:val3

    The webclient includes some unfinished support for this field to be used to
    create circular annotations with YouTube video links. If your WTML file will
    not be viewed in the webclient, you can use this field to convey arbitrary
    textual data to the WWT Web Engine JavaScript/TypeScript layer.

    """

    msr_community_id = Int(0).tag(xml=XmlSer.attr('MSRCommunityId'))
    """The ID number of the WWT Community that this content came from."""

    msr_component_id = Int(0).tag(xml=XmlSer.attr('MSRComponentId'))
    """The ID number of this content item on the WWT Communities system."""

    permission = Int(0).tag(xml=XmlSer.attr('Permission'))
    "TBD."

    xmeta = Instance(
        Namespace,
        args = (),
        help = 'XML metadata - a namespace object for attaching arbitrary text to serialize',
    ).tag(xml=XmlSer.ns_to_attr('X'))

    def _tag_name(self):
        return 'Place'

    def mutate_urls(self, mutator):
        if self.thumbnail:
            self.thumbnail = mutator(self.thumbnail)

        if self.background_image_set:
            self.background_image_set.mutate_urls(mutator)

        if self.foreground_image_set:
            self.foreground_image_set.mutate_urls(mutator)

        if self.image_set:
            self.image_set.mutate_urls(mutator)

    def as_imageset(self):
        """Return an ImageSet for this place if one is defined.

        Returns
        -------
        Either :class:`wwt_data_formats.imageset.ImageSet` or None.

        Notes
        -----
        If the :attr:`foreground_image_set` of this :class:`Place` is not
        None, it is returned. Otherwise, if its :attr:`image_set` is not
        None, that is returned. Otherwise, None is returned.

        """
        if self.foreground_image_set is not None:
            return self.foreground_image_set
        return self.image_set



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020-2021 the .NET Foundation
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import numpy as np
import numpy.testing as nt
import pytest
from xml.etree import ElementTree as etree

from . import assert_xml_trees_equal
from .. import imageset, enums, stringify_xml_doc, write_xml_doc


def test_basic_xml():
    expected_str = '''
<ImageSet MSRCommunityId="0" MSRComponentId="0" Permission="0"
          BandPass="Gamma" BaseDegreesPerTile="0.1" BaseTileLevel="1"
          BottomsUp="True" CenterX="1.234" CenterY="-0.31415"
          DataSetType="Planet" ElevationModel="False" FileType=".PNG" Generic="False"
          MeanRadius="0.0" Name="Test name"
          OffsetX="100.1" OffsetY="100.2" Projection="SkyImage"
          Rotation="5.4321" Sparse="False" StockSet="False" TileLevels="4"
          Url="http://example.org/{0}" WidthFactor="2">
  <Credits>Escaping &amp; Entities</Credits>
  <CreditsUrl>https://example.org/credits</CreditsUrl>
  <Description>Escaping &lt;entities&gt;</Description>
  <ThumbnailUrl>https://example.org/thumbnail.jpg</ThumbnailUrl>
</ImageSet>
'''
    expected_xml = etree.fromstring(expected_str)

    imgset = imageset.ImageSet()
    imgset.data_set_type = enums.DataSetType.PLANET
    imgset.name = 'Test name'
    imgset.url = 'http://example.org/{0}'
    imgset.width_factor = 2
    imgset.base_tile_level = 1
    imgset.tile_levels = 4
    imgset.base_degrees_per_tile = 0.1
    imgset.file_type = '.PNG'
    imgset.bottoms_up = True
    imgset.projection = enums.ProjectionType.SKY_IMAGE
    imgset.center_x = 1.234
    imgset.center_y = -0.31415
    imgset.offset_x = 100.1
    imgset.offset_y = 100.2
    imgset.rotation_deg = 5.4321
    imgset.band_pass = enums.Bandpass.GAMMA
    imgset.sparse = False
    imgset.credits = 'Escaping & Entities'
    imgset.credits_url = 'https://example.org/credits'
    imgset.thumbnail_url = 'https://example.org/thumbnail.jpg'
    imgset.description = 'Escaping <entities>'

    observed_xml = imgset.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)


def test_wcs_1():
    expected_str = '''
<ImageSet BandPass="Visible" BaseDegreesPerTile="4.870732233333334e-05"
          BaseTileLevel="0" BottomsUp="True" CenterX="83.633083"
          CenterY="22.0145" DataSetType="Sky" ElevationModel="False"
          FileType=".png" Generic="False" MeanRadius="0.0" MSRCommunityId="0"
          MSRComponentId="0" OffsetX="1503.3507831457316"
          OffsetY="1479.3005935660037" Permission="0" Projection="SkyImage"
          Rotation="-179.70963521481" Sparse="True" StockSet="False"
          TileLevels="0" WidthFactor="2">
</ImageSet>
'''
    expected_xml = etree.fromstring(expected_str)

    wcs_keywords = {
        'CTYPE1': 'RA---TAN',
        'CTYPE2': 'DEC--TAN',
        'CRVAL1': 83.633083,
        'CRVAL2': 22.0145,
        'PC1_1': 0.9999871586199364,
        'PC1_2': 0.005067799840785529,
        'PC2_1': -0.005067799840785529,
        'PC2_2': 0.9999871586199364,
        'CRPIX1': 1503.8507831457316,
        'CRPIX2': 1479.8005935660037,
        'CDELT1': -4.870732233333334e-05,
        'CDELT2': 4.870732233333334e-05,
    }

    imgset = imageset.ImageSet()
    imgset.set_position_from_wcs(wcs_keywords, 3000, 3000)

    observed_xml = imgset.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)

    wcs_roundtrip = imgset.wcs_headers_from_position()

    # Postprocess CD into PC/CDELT using the usual normalization

    det = wcs_roundtrip['CD1_1'] * wcs_roundtrip['CD2_2'] - wcs_roundtrip['CD1_2'] * wcs_roundtrip['CD2_1']
    if det < 0:
        cd_sign = -1
    else:
        cd_sign = 1

    wcs_roundtrip['CDELT1'] = np.sqrt(wcs_roundtrip['CD1_1']**2 + wcs_roundtrip['CD1_2']**2) * cd_sign
    wcs_roundtrip['CDELT2'] = np.sqrt(wcs_roundtrip['CD2_1']**2 + wcs_roundtrip['CD2_2']**2)
    wcs_roundtrip['PC1_1'] = wcs_roundtrip['CD1_1'] / wcs_roundtrip['CDELT1']
    wcs_roundtrip['PC1_2'] = wcs_roundtrip['CD1_2'] / wcs_roundtrip['CDELT1']
    wcs_roundtrip['PC2_1'] = wcs_roundtrip['CD2_1'] / wcs_roundtrip['CDELT2']
    wcs_roundtrip['PC2_2'] = wcs_roundtrip['CD2_2'] / wcs_roundtrip['CDELT2']

    for hn in 'CD1_1 CD1_2 CD2_1 CD2_2'.split():
        del wcs_roundtrip[hn]

    # OK, now we can compare again

    for kw in wcs_roundtrip.keys():
        expected = wcs_keywords[kw]
        observed = wcs_roundtrip[kw]

        if kw in ('CTYPE1', 'CTYPE2'):
            assert expected == observed
        else:
            nt.assert_almost_equal(expected, observed)


def test_wcs_ok_matrices():
    base_keywords = {
        'CTYPE1': 'RA---TAN',
        'CTYPE2': 'DEC--TAN',
        'CRVAL1': 0,
        'CRVAL2': 0,
        'CRPIX1': 0,
        'CRPIX2': 0,
    }

    case1 = {
        'CD1_1': -0.05,
        'CD1_2': -0.3,
        'CD2_1': -0.3,
        'CD2_2': 0.05,
    }

    case2 = {
        'CD1_1': -0.05,
        'CD1_2': -0.3,
        'CD2_1': 0.3,
        'CD2_2': -0.05,
    }

    for these_kws in (case1, case2):
        all_kws = dict(base_keywords)
        all_kws.update(these_kws)

        imgset = imageset.ImageSet()
        imgset.set_position_from_wcs(all_kws, 1000, 1000)
        roundtrip_kws = imgset.wcs_headers_from_position()

        # Postprocess if needed

        if 'CDELT1' in these_kws:
            det = roundtrip_kws['CD1_1'] * roundtrip_kws['CD2_2'] - roundtrip_kws['CD1_2'] * roundtrip_kws['CD2_1']
            if det < 0:
                cd_sign = -1
            else:
                cd_sign = 1

            roundtrip_kws['CDELT1'] = np.sqrt(roundtrip_kws['CD1_1']**2 + roundtrip_kws['CD1_2']**2) * cd_sign
            roundtrip_kws['CDELT2'] = np.sqrt(roundtrip_kws['CD2_1']**2 + roundtrip_kws['CD2_2']**2)
            roundtrip_kws['PC1_1'] = roundtrip_kws['CD1_1'] / roundtrip_kws['CDELT1']
            roundtrip_kws['PC1_2'] = roundtrip_kws['CD1_2'] / roundtrip_kws['CDELT1']
            roundtrip_kws['PC2_1'] = roundtrip_kws['CD2_1'] / roundtrip_kws['CDELT2']
            roundtrip_kws['PC2_2'] = roundtrip_kws['CD2_2'] / roundtrip_kws['CDELT2']

            for hn in 'CD1_1 CD1_2 CD2_1 CD2_2'.split():
                del roundtrip_kws[hn]

        # OK, now we can compare

        for kw, expected in these_kws.items():
            observed = roundtrip_kws[kw]
            nt.assert_almost_equal(expected, observed)


def test_wcs_bad_matrices():
    base_keywords = {
        'CTYPE1': 'RA---TAN',
        'CTYPE2': 'DEC--TAN',
        'CRVAL1': 0,
        'CRVAL2': 0,
        'CRPIX1': 0,
        'CRPIX2': 0,
    }

    case1 = {
        'CD1_1': 0.05,
        'CD1_2': 0.3,
        'CD2_1': 0.3,
        'CD2_2': 0.05,
    }

    case2 = {
        'CD1_1': -0.05,
        'CD1_2': -0.3,
        'CD2_1': -0.3,
        'CD2_2': -0.05,
    }

    case3 = {
        'CD1_1': -0.05,
        'CD1_2': -0.3,
        'CD2_1': 0.3,
        'CD2_2': 0.05,
    }

    case4 = {
        'CD1_1': -0.05,
        'CD1_2': 0.3,
        'CD2_1': -0.3,
        'CD2_2': 0.05,
    }

    for these_kws in (case1, case2, case3, case4):
        all_kws = dict(base_keywords)
        all_kws.update(these_kws)

        imgset = imageset.ImageSet()

        with pytest.raises(ValueError):
            imgset.set_position_from_wcs(all_kws, 1000, 1000)


def test_wcs_tiled_topdown():
    """
    If you're setting WCS for a tiled image, it must have negative (JPEG,
    top-down) parity -- bottoms-up tiled images won't render

    """
    keywords = {
        'CTYPE1': 'RA---TAN',
        'CTYPE2': 'DEC--TAN',
        'CRVAL1': 0,
        'CRVAL2': 0,
        'CRPIX1': 0,
        'CRPIX2': 0,
        'CD1_1': 0.05,
        'CD1_2': 0.3,
        'CD2_1': 0.3,
        'CD2_2': -0.05,
    }

    imgset = imageset.ImageSet()
    imgset.tile_levels = 5

    with pytest.raises(Exception):
        imgset.set_position_from_wcs(keywords, 1000, 1000)

    # flip parity, as far as CD is concerned ...
    keywords['CD2_1'] *= -1
    keywords['CD2_2'] *= -1
    # This should now work:
    imgset.set_position_from_wcs(keywords, 1000, 1000)


def test_misc_ser():
    expected_str = '''
<ImageSet BandPass="Visible" BaseDegreesPerTile="0.0" BaseTileLevel="0"
          BottomsUp="False" CenterX="0.0" CenterY="0.0" DataSetType="Sky" ElevationModel="False"
          FileType=".png" Generic="False" MeanRadius="0.0" MSRCommunityId="0" MSRComponentId="0"
          OffsetX="0.0" OffsetY="0.0" Permission="0" Projection="SkyImage"
          Rotation="0.0" Sparse="True" StockSet="False" TileLevels="0"
          Url="http://example.com/unspecified" WidthFactor="2" />
'''
    expected_xml = etree.fromstring(expected_str)

    imgset = imageset.ImageSet()
    imgset.url = 'http://example.com/unspecified'

    observed_xml = imgset.to_xml()
    assert_xml_trees_equal(expected_xml, observed_xml)

    expected_text = stringify_xml_doc(expected_xml)
    observed_text = imgset.to_xml_string()
    assert observed_text == expected_text

    from io import BytesIO

    expected_bio = BytesIO()
    observed_bio = BytesIO()
    write_xml_doc(expected_xml, dest_stream=expected_bio, dest_wants_bytes=True)
    imgset.write_xml(observed_bio, dest_wants_bytes=True)
    assert observed_bio.getvalue() == expected_bio.getvalue()
    assert b'encoding=\'UTF-8\'' in observed_bio.getvalue()



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2021 the .NET Foundation
# Licensed under the MIT License.

"""
An image, possibly tiled, for display in WWT.
"""

from __future__ import absolute_import, division, print_function

__all__ = '''
ImageSet
'''.split()

import math
from traitlets import Bool, Float, Int, Unicode, UseEnum
from xml.etree import ElementTree as etree

from . import LockedXmlTraits, XmlSer
from .abcs import UrlContainer
from .enums import Bandpass, DataSetType, ProjectionType


class ImageSet(LockedXmlTraits, UrlContainer):
    """
    A WWT imagery dataset.

    Instances of this class express WWT imagery datasets and their spatial
    positioning. Imagesets are exposed to the WWT engine through their XML
    serialization in a WTML :class:`~wwt_data_formats.folder.Folder`. The engine
    can be instructed to load such a folder, making its imagesets available for
    rendering.
    """

    data_set_type = UseEnum(DataSetType, default_value=DataSetType.SKY).tag(xml=XmlSer.attr('DataSetType'))
    """The renderer mode to which these data apply.

    Possible values are ``"Earth"``, ``"Planet"``, ``"Sky"``, ``"Panorama"``,
    ``"SolarSystem"``, and ``"Sandbox"``.

    """
    reference_frame = Unicode('').tag(xml=XmlSer.attr('ReferenceFrame'))
    """TBD."""

    name = Unicode('').tag(xml=XmlSer.attr('Name'))
    """A name used to refer to this imageset.

    Various parts of the WWT internals reference imagesets by this name, so
    it should be distinctive.

    """
    url = Unicode('').tag(xml=XmlSer.attr('Url'))
    """The URL of the image data.

    Either a URL or a URL template. URLs that are exposed to the engine should
    be absolute and use the ``http://`` protocol (the web engine will rewrite
    them to HTTPS if needed). The ``wwtdatatool`` program that comes with this
    package provides some helpful utilities to allow data-processing to use relative URLs. TODO: more
    details.

    """
    alt_url = Unicode('').tag(xml=XmlSer.attr('AltUrl'))
    """An alternative URL that provided the data.

    If provided and the Windows client attempts to load an imageset using this
    alternative URL, that imageset will be replaced by this one. This provides a
    mechanism for superseding old imagesets with improved versions.

    """
    dem_url = Unicode('').tag(xml=XmlSer.attr('DemUrl'))
    """The URL of the DEM data.

    Either a URL or a URL template. TODO: details

    """
    width_factor = Int(2).tag(xml=XmlSer.attr('WidthFactor'))
    """This is a legacy parameter. Leave it at 2."""

    base_tile_level = Int(0).tag(xml=XmlSer.attr('BaseTileLevel'))
    """The level of the highest (coarsest-resolution) tiling available.

    This should be zero except for special circumstances.

    """
    quad_tree_map = Unicode('').tag(xml=XmlSer.attr('QuadTreeMap'))
    """TBD."""

    tile_levels = Int(0).tag(xml=XmlSer.attr('TileLevels'))
    """The number of levels of tiling.

    Should be zero for untiled images (``projection =
    ProjectionType.SkyImage``).

    For tiled images (``projection = ProjectionType.Tan``), an image with
    ``tile_levels = 1`` has been broken into four tiles, each 256x256 pixels.
    For ``tile_levels = 2``, there are sixteen tiles, and the padded height of
    the tiled area is ``256 * 2**2 = 1024`` pixels. Image with dimensions of
    2048 pixels or smaller do not need to be tiled, so if this parameter is
    nonzero it will usually be 4 or larger.

    """
    base_degrees_per_tile = Float(0.0).tag(xml=XmlSer.attr('BaseDegreesPerTile'))
    """The angular scale of the image.

    For untiled images, this is the pixel scale: the number of degrees per pixel
    in the vertical direction. Non-square pixels are not supported.

    For tiled images, this is the angular height of the image, in degrees, after
    its dimensions have been padded out to the next largest power of 2 for
    tiling purposes. If a square image is 1200 pixels tall and has a height of
    0.016 deg, the padded height would be 2048 pixels and this parameter should
    be set to 0.016 * 2048 / 1200 = 0.0273.

    """
    file_type = Unicode('.png').tag(xml=XmlSer.attr('FileType'))
    """
    The extension(s) of the image file(s) in this set.

    In the simplest case, this field will contain an image filetype extension
    including the leading period, such as ``.jpeg`` or ``.png``. Some datasets
    in the wild lack the leading period: they have just ``png`` or something
    similar. The value ``.auto`` is also used in some cases, which can be OK
    because often WWT doesn't actually use this field for any particular
    purpose.

    Some datasets, like HiPS imagery, provide multiple filetypes simultaneously.
    These can be expressed by including several filename extensions separated by
    spaces. For instance, ``png jpeg fits``. The existing WTML records that
    support multiple filetypes do not include any leading periods, but clients
    should be prepared for them to be present.

    Imagesets to be rendered as FITS data *must* have the exact value ``.fits``
    for this field. If multiple filetypes are specified, the special
    FITS-rendering machinery will not be invoked. This is true for both single
    FITS files and tiled FITS imagesets, including HiPS FITS datasets.

    A supported filetype extension of ``tsv`` (or ``.tsv``) means that this
    "imageset" actually contains a HiPS progressive catalog, not bitmap imagery.
    Imageset records should not intermix image-type and catalog-type filetypes.
    (We don't know if there are any examples in the wild of HiPS datasets that
    claim to contain both kinds of data.)
    """

    bottoms_up = Bool(False).tag(xml=XmlSer.attr('BottomsUp'))
    """
    The parity of the image's projection on the sky.

    For untiled (``projection = SkyImage``) images, this flag defines the
    image's parity, which basically sets whether the image needs to be flipped
    during rendering. This field should be False for typical RGB color images
    that map onto the sky as if you had taken them with a digital camera. For
    these images, the first row of image data is at the top of the image at zero
    rotation. For typical FITS files, on the other hand, the first row of image
    data is at the bottom of the image, which results in a parity inversion. In
    these cases, the ``bottoms_up`` flag should be True (hence its name). In the
    terminology of `Astrometry.Net
    <https://astroquery.readthedocs.io/en/latest/astrometry_net/astrometry_net.html#parity>`_,
    ``bottoms_up = False`` corresponds to negative parity, and ``bottoms_up =
    True`` corresponds to positive parity.

    The effect of setting this flag to True is to effectively flip the image and
    its coordinate system left-to-right. For a ``bottoms_up = False`` image with
    :attr:`offset_x`, :attr:`offset_y`, and :attr:`rotation_deg` all zero, the
    lower-left corner of the image lands at the :attr:`center_x` and
    :attr:`center_y`, and positive rotations rotate the image counter-clockwise
    around that origin. If you take the same image and make ``bottoms_up =
    True``, the image will appear to have been flipped left-to-right, the
    lower-*right* corner of the image will land at the coordinate center, and
    positive rotations will rotate it *clockwise* around that origin. In both
    cases, positive values of :attr:`offset_x` and :attr:`offset_y` move the
    center of the image closer to the coordinate center, but when ``bottoms_up =
    False``, this means that the image is moving down and left, and when
    ``bottoms_up = True`` this means that the image is moving down and right.

    For tiled images (``projection = Tan``), this field must be false. If it is
    true, the imageset won't render.
    """

    projection = UseEnum(
        ProjectionType,
        default_value = ProjectionType.SKY_IMAGE
    ).tag(xml=XmlSer.attr('Projection'))
    """The type of projection used to place this image on the sky.

    For untiled images, this should be "SkyImage". For tiled images, it should
    be "Tan". The :meth:`set_position_from_wcs` method will set this value
    appropriately based on :attr:`tile_levels`.

    """
    center_x = Float(0.0).tag(xml=XmlSer.attr('CenterX'))
    """The horizontal location of the center of the image’s projection
    coordinate system.

    For sky images, this is a right ascension in degrees. Note that this
    parameter just helps to define a coordinate system; it does not control how
    the actual image data are placed onto that coordinate system. The
    :attr:`offset_x` and :attr:`offset_y` parameters do that.

    """
    center_y = Float(0.0).tag(xml=XmlSer.attr('CenterY'))
    """The vertical location of the center of the image’s projection coordinate
    system.

    For sky images, this is a declination in degrees. Note that this parameter
    just helps to define a coordinate system; it does not control how the actual
    image data are placed onto that coordinate system. The :attr:`offset_x` and
    :attr:`offset_y` parameters do that.

    """
    offset_x = Float(0.0).tag(xml=XmlSer.attr('OffsetX'))
    """
    The horizontal positioning of the image relative to its projection
    coordinate system.

    For untiled sky images with :attr:`bottoms_up` false, the image is by
    default positioned such that its lower left corner lands at the center of
    the projection coordinate system (namely, :attr:`center_x` and
    :attr:`center_y`). The offset is measured in pixels and moves the image
    leftwards. Therefore, ``offset_x = image_width / 2`` places the horizontal
    center of the image at ``center_x``. This parameter is therefore analogous
    to the WCS keyword ``CRVAL1``.

    For untiled sky images where :attr:`bottoms_up` is true, the X coordinate
    system has been mirrored. Therefore when this field is zero, the lower
    *right* corner of the image will land at the center of the projection
    coordinate system, and positive values will move the image to the right.

    For tiled sky images, the offset is measured in *degrees*, and a value of
    zero means that the *center* of the image lands at the center of the
    projection coordinate system. Increasingly positive values move the image to
    the right.

    As per the usual practice, offsets are always along the horizontal axis of
    the image in question, regardless of its :attr:`rotation <rotation_deg>` on
    the sky.

    """
    offset_y = Float(0.0).tag(xml=XmlSer.attr('OffsetY'))
    """The vertical positioning of the image relative to its projection
    coordinate system.

    For untiled sky images with :attr:`bottoms_up` false, the image is by
    default positioned such that its lower left corner lands at the center of
    the projection coordinate system (namely, :attr:`center_x` and
    :attr:`center_y`). The offset is measured in pixels and moves the image
    downwards. Therefore, ``offset_y = image_height / 2`` places the vertical
    center of the image at ``center_y``. This parameter is therefore analogous
    to the WCS keyword ``CRVAL2``.

    For untiled sky images where :attr:`bottoms_up` is true, the X coordinate
    system has been mirrored but the Y coordinate system is the same. Therefore
    when this field is zero, the lower *right* corner of the image will land at
    the center of the projection coordinate system, but positive values will
    still move the image downwards.

    For tiled sky images, the offset is measured in *degrees*, and a value of
    zero means that the *center* of the image lands at the center of the
    projection coordinate system. Increasingly positive values move the image
    upwards.

    As per the usual practice, offsets are always along the vertical axis of the
    image in question, regardless of its :attr:`rotation <rotation_deg>` on the
    sky.

    """
    rotation_deg = Float(0.0).tag(xml=XmlSer.attr('Rotation'))
    """
    The rotation of image’s projection coordinate system, in degrees.

    For sky images with :attr:`bottoms_up` false, this is East from North, i.e.
    counterclockwise. If :attr:`bottoms_up` is true (only allowed for untiled
    images), the image coordinate system is mirrored, and positive rotations
    rotate the image *clockwise* relative to the sky.
    """

    band_pass = UseEnum(
        Bandpass,
        default_value = Bandpass.VISIBLE
    ).tag(xml=XmlSer.attr('BandPass'))
    """The bandpass of the image data."""

    sparse = Bool(True).tag(xml=XmlSer.attr('Sparse'))
    """TBD."""

    elevation_model = Bool(False).tag(xml=XmlSer.attr('ElevationModel'))
    """TBD."""

    stock_set = Bool(False).tag(xml=XmlSer.attr('StockSet'))
    """TBD."""

    generic = Bool(False).tag(xml=XmlSer.attr('Generic'))
    """TBD."""

    mean_radius = Float(0.0).tag(xml=XmlSer.attr('MeanRadius'))
    """TBD."""

    credits = Unicode('').tag(xml=XmlSer.text_elem('Credits'))
    """Textual credits for the image originator."""

    credits_url = Unicode('').tag(xml=XmlSer.text_elem('CreditsUrl'))
    """A URL giving the source of the image or more information about its creation."""

    thumbnail_url = Unicode('').tag(xml=XmlSer.text_elem('ThumbnailUrl'))
    """A URL to a standard WWT thumbnail representation of this imageset."""

    description = Unicode('').tag(xml=XmlSer.text_elem('Description'))
    """
    A textual description of the imagery.

    This field is referenced a few times in the original WWT documentation, but
    is not actually implemented. The ``Place.description`` field is at least
    loaded from the XML.
    """

    msr_community_id = Int(0).tag(xml=XmlSer.attr('MSRCommunityId'))
    """The ID number of the WWT Community that this content came from."""

    msr_component_id = Int(0).tag(xml=XmlSer.attr('MSRComponentId'))
    """The ID number of this content item on the WWT Communities system."""

    permission = Int(0).tag(xml=XmlSer.attr('Permission'))
    "TBD."

    def _tag_name(self):
        return 'ImageSet'

    def mutate_urls(self, mutator):
        if self.url:
            self.url = mutator(self.url)

        if self.dem_url:
            self.dem_url = mutator(self.dem_url)

        if self.credits_url:
            self.credits_url = mutator(self.credits_url)

        if self.thumbnail_url:
            self.thumbnail_url = mutator(self.thumbnail_url)

    def set_position_from_wcs(self, headers, width, height, place=None, fov_factor=1.7):
        """
        Set the positional information associated with this imageset to match a
        set of WCS headers.

        Parameters
        ----------
        headers : :class:`~astropy.io.fits.Header` or string-keyed dict-like

            A set of FITS-like headers including WCS keywords such as ``CRVAL1``.

        width : positive integer

            The width of the image associated with the WCS, in pixels.

        height : positive integer

            The height of the image associated with the WCS, in pixels.

        place : optional :class:`~wwt_data_formats.place.Place`

            If specified, the centering and zoom level of the
            :class:`~wwt_data_formats.place.Place` object will be set to match
            the center and size of this image.

        fov_factor : optional

            If *place* is provided, its zoom level will be set so that the angular
            height of the client viewport is this factor times the angular height of
            the image. The default is 1.7.

        Returns
        -------
        **self**

            For convenience in chaining function calls.

        Notes
        -----
        Certain of the ImageSet parameters take on different meanings depending
        on whether the image in question is a tiled "study" or not. This method
        will alter its behavior depending on whether the :attr:`tile_levels`
        attribute is greater than zero. **Make sure that this parameter has
        acquired its final value before calling this function.**

        For the time being, the WCS must be equatorial using the gnomonic
        (``TAN``) projection.

        Required keywords in *headers* are:

        - ``CTYPE1`` and ``CTYPE2``
        - ``CRVAL1`` and ``CRVAL2``
        - ``CRPIX1`` and ``CRPIX2``
        - Either:
            - ``CD1_1``, ``CD2_2`` (preferred)
            - ``CDELT1``, ``CDELT2``, ``PC1_1``, and ``PC1_2``; or

        If present ``PC1_2``, ``PC2_1``, ``CD1_2``, and/or ``CD2_1`` are used.
        If absent, they are assumed to be zero.

        This routine assumes that the WCS coordinates have the correct parity
        for the data that they describe. If these WCS coordinates describe a
        JPEG-like image, the parity of the coordinates should be negative
        ("top-down"), which means that the determinant of the CD matrix should
        have a *positive* sign. If these coordinates describe a FITS-like image,
        the parity should be positive or "bottoms-up", with a negative
        determinant of the CD matrix. If the image in question is tiled, the
        parity must be top-down, in the sense the bottoms-up tiled imagery just
        won't render in the engine. There are some CD matrices that can't be
        expressed in WWT's formalism (rotation, scale, parity) and this method
        will do its best to detect and reject them.
        """

        if headers['CTYPE1'] != 'RA---TAN' or headers['CTYPE2'] != 'DEC--TAN':
            raise ValueError('WCS coordinates must be in an equatorial/TAN projection')

        # Figure out the stuff we need from the headers.

        ra_deg = headers['CRVAL1']
        dec_deg = headers['CRVAL2']

        # In FITS/WCS, pixel coordinates are 1-based and integer pixel
        # coordinates land on pixel centers. Therefore in standard FITS
        # orientation, where the "first" pixel is at the lower-left, the
        # lower-left corner of the image has pixel coordinate (0.5, 0.5). For
        # the WWT offset parameters, the lower-left corner of the image has
        # coordinate (0, 0).
        refpix_x = headers['CRPIX1'] - 0.5
        refpix_y = headers['CRPIX2'] - 0.5

        if 'CD1_1' in headers:
            cd1_1 = headers['CD1_1']
            cd2_2 = headers['CD2_2']
            cd1_2 = headers.get('CD1_2', 0.0)
            cd2_1 = headers.get('CD2_1', 0.0)
        else:
            # older PC/CDELT form -- note that we're using two additional
            # numbers to express the same information.
            d1 = headers['CDELT1']
            d2 = headers['CDELT2']
            cd1_1 = d1 * headers.get('PC1_1', 1.0)
            cd2_2 = d2 * headers.get('PC2_2', 1.0)
            cd1_2 = d1 * headers.get('PC1_2', 0.0)
            cd2_1 = d2 * headers.get('PC2_1', 0.0)

        cd_det = cd1_1 * cd2_2 - cd1_2 * cd2_1

        if cd_det < 0:
            cd_sign = -1

            if self.tile_levels > 0:
                raise Exception('WCS for tiled imagery must have top-down/negative/JPEG_like parity')
        else:
            cd_sign = 1

        # Given how WWT implements its rotation coordinates, this expression
        # turns out to give us the correct value for different rotations and
        # parities:

        rot_rad = math.atan2(-cd_sign * cd1_2, -cd2_2)

        # We can only express square pixels with a rotation and a potential
        # parity flip. Do some cross-checks to ensure that the input matrix can
        # be well-approximated this way. There's probably a smarter
        # linear-algebra way to do this.

        TOL = 0.05

        if not abs(cd_det) > 0:
            raise ValueError('determinant of the CD matrix is not positive')

        scale_x = math.sqrt(cd1_1**2 + cd1_2**2)
        scale_y = math.sqrt(cd2_1**2 + cd2_2**2)

        if abs(scale_x - scale_y) / (scale_x + scale_y) > TOL:
            raise ValueError('WWT cannot express non-square pixels, which this WCS has')

        if abs((cd1_1 - cd_sign * cd2_2) / cd_det) > TOL:
            raise ValueError('WWT cannot express this CD matrix (1)')

        if abs((cd2_1 + cd_sign * cd1_2) / cd_det) > TOL:
            raise ValueError('WWT cannot express this CD matrix (2)')

        # This is our best effort to make sure that the view centers on the
        # center of the image.

        try:
            from astropy.wcs import WCS
        except:
            center_ra_deg = ra_deg
            center_dec_deg = dec_deg
        else:
            wcs = WCS(headers)
            # The WCS object uses 0-based indices where the integer coordinates
            # land on pixel centers. Therefore the dead center of the image has
            # pixel coordinates as below. For instance, in an image 2 pixels
            # wide, the horizontal center is where the two pixels touch, which
            # has an X coordinate of 0.5.
            center = wcs.pixel_to_world((width - 1) / 2, (height - 1) / 2)
            center_ra_deg = center.ra.deg
            center_dec_deg = center.dec.deg

        # Now, assign the fields

        self.data_set_type = DataSetType.SKY
        self.width_factor = 2
        self.center_x = ra_deg
        self.center_y = dec_deg
        self.rotation_deg = rot_rad * 180 / math.pi

        if self.tile_levels > 0:  # are we tiled?
            self.projection = ProjectionType.TAN
            self.offset_x = (width / 2 - refpix_x) * scale_x
            self.offset_y = (refpix_y - height / 2) * scale_y
            self.base_degrees_per_tile = scale_y * 256 * 2**self.tile_levels
        else:
            self.projection = ProjectionType.SKY_IMAGE
            self.offset_x = refpix_x
            self.offset_y = refpix_y
            self.base_degrees_per_tile = scale_y
            self.bottoms_up = (cd_sign == -1)

        if place is not None:
            place.data_set_type = DataSetType.SKY
            place.rotation_deg = 0.  # I think this is better than propagating the image rotation?
            place.ra_hr = center_ra_deg / 15.
            place.dec_deg = center_dec_deg
            # It is hardcoded that in sky mode, zoom_level = height of client FOV * 6.
            place.zoom_level = height * scale_y * fov_factor * 6

        return self


    def wcs_headers_from_position(self):
        """
        Compute a set of WCS headers for this ImageSet's positional information.

        Returns
        -------
        A string-keyed dict-like containing FITS/WCS header keywords such as
        ``CTYPE1``, ``CRPIX1``, etc.

        Notes
        -----
        At the moment, this function only works for ImageSets with a projection
        type of ``SKY_IMAGE``. Support for other projections *might* be added
        later, if the need arises. Note, however, that tiled images have their
        sizes adjusted to be powers of 2, and the "actual" size of the source
        imagery is not necessarily preserved.
        """

        rv = {
            'CTYPE1': 'RA---TAN',
            'CTYPE2': 'DEC--TAN',
            'CRVAL1': self.center_x,
            'CRVAL2': self.center_y,
        }

        if self.projection != ProjectionType.SKY_IMAGE:
            raise NotImplementError('wcs_headers_from_position() only works if projection=SKY_IMAGE')

        rv['CRPIX1'] = self.offset_x + 0.5
        rv['CRPIX2'] = self.offset_y + 0.5

        # The WWT rotation angle is 180 degrees away from the usual angle
        # that you would use for a rotation matrix, which is why we negate
        # these trig values:
        c = -math.cos(self.rotation_deg * math.pi / 180)
        s = -math.sin(self.rotation_deg * math.pi / 180)
        parity = -1 if self.bottoms_up else 1

        # | CD1_1 CD1_2 | = scale * | p 0 | * |  cos(theta) sin(theta) |
        # | CD2_1 CD2_2 |           | 0 1 |   | -sin(theta) cos(theta) |

        rv['CD1_1'] = c * self.base_degrees_per_tile * parity
        rv['CD1_2'] = s * self.base_degrees_per_tile * parity
        rv['CD2_1'] = -s * self.base_degrees_per_tile
        rv['CD2_2'] = c * self.base_degrees_per_tile

        return rv
